#ifndef __CURSOR_CPP
#define __CURSOR_CPP

#include <iostream>
#include "getch.h"

/**
 * @brief Set of tools to manipulate cursor in terminal.
 */
namespace cursor {
/**
 * @brief Represents cursor coordinates.
 */
struct position {
    short x = 0; // column
    short y = 0; // line
};

/**
 * @brief Get current cursor position.
 * 
 * @return position the current position.
 */
position getPosition() {
    position returnVal;
    std::cout << "\033[6n";

    bool readSemicolon = false;
    bool stillReading = true;

    while (stillReading) {
        char read = getch();

        // \033[{line};{column}R
        switch (read)
        {
        case '\033':
        case '[':
            break;
        case ';':
            readSemicolon = true;
            break;
        case 'R':
            stillReading = false;
            break;
        default:
            // if character read is one of: 0123456789
            if (read >= 0x30 & read <= 0x39) {
                read -= 0x30;
                if (readSemicolon) // if reading x pos
                    returnVal.x = 10 * returnVal.x + read;
                else
                    returnVal.y = 10 * returnVal.y + read;
            }
            else {std::cout << (int)read << ",";}
            break;
        }
    }
    return returnVal;
}

/**
 * @brief Set the position of a cursor in terminal.
 * 
 * @param x the column to which to set.
 * @param y the line to which to set.
 */
void setPosition(const short& x, const short& y) {
    std::cout << "\033[" << y << ";" << x << "H";
}

/**
 * @brief Set the position of a cursor in terminal.
 * 
 * @param pos the position to which to set.
 */
void setPosition(const position& pos) {
    std::cout << "\033[" << pos.y << ";" << pos.x << "H";
}

/**
 * @brief Sets the column of a cursor.
 * 
 * @param _x The column to which to set.
 */
void setX(const short& _x) { std::cout << "\033[" << _x << "G"; }

/**
 * @brief Set the row of a cursor.
 * 
 * @param _y The row to which to set.
 */
void setY(const short& _y) { setPosition(getPosition().x, _y); }

/**
 * @brief Requests terminal to save position.
 */
void savePosition() { std::cout << "\0337"; }

/**
 * @brief Requests termianl to restore saved position.
 */
void restorePosition() { std::cout << "\0338"; }

/**
 * @brief moves cursor 1 line up
 */
void up() { std::cout << "\033[1A"; }

/**
 * @brief represents pressed key.
 * 
 */
struct key {
    unsigned char code = 0;
    unsigned char special = 0;
};

/**
 * @brief gets keyboard characker.
 * 
 * @return key the pressed key.
 */
key getchar() {
    key returnval;
    unsigned char input = getch();
    switch (input) {
    case 0: // special
        returnval.special = 1;
        returnval.code = getch();
        break;
    case 224: // non numpad navigation keys
        returnval.special = 2;
        returnval.code = getch();
        break;
    default:
        returnval.code = input;
        break;
    }
    return returnval;
}

}; // namespace Cursor

#endif
